/*****************************************************************************\
 *                                                                           *
 * Xrtti.h                                                                   *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
\*****************************************************************************/


#ifndef XRTTI_H
#define XRTTI_H

#include <typeinfo>

/** **************************************************************************
 * Everything which follows is in the Xrtti C++ namespace.
 ************************************************************************** **/

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif

/** **************************************************************************
 * Fundamental Types
 ************************************************************************** **/

/**
 * The following must be defined before this header file is included:
 * XRTTI_U32_ARCH  --  unsigned 32 bit integer
 * XRTTI_S32_ARCH  --  signed 32 bit integer
 **/

typedef XRTTI_U32_ARCH     U32;       /**< 32-bit unsigned integer */
typedef XRTTI_S32_ARCH     S32;       /**< 32-bit signed integer */


/** **************************************************************************
 * Forward declaration of classes and structures
 ************************************************************************** **/

class Argument;
class AutoRegister;                /**< AutoRegister class, opaque type */
class Array;
class Base;
class Class;
class Complete;
class Constructor;
class ConstructorSignature;
class Context;
class Destructor;
class DestructorSignature;
class Enumeration;
class EnumerationValue;
class Field;
typedef struct Generated Generated; /**< Generated class, opaque type */
class Incomplete;
class Method;
class MethodSignature;
class Namespace;
class Pointer;
class Structure;
class Struct;
class Type;
class TypeEnumeration;
class TypeFunction;
class TypeStructure;
class TypeUnion;
class Union;


/** **************************************************************************
 * AccessType defines the access types of C++, which are used to define how
 * code may access classes, structs, methods, fields, and enumerations
 ************************************************************************** **/
typedef enum AccessType
{
    AccessType_Public,
    AccessType_Protected,
    AccessType_Private
} AccessType;


/** **************************************************************************
 * Instances of the Value union are passed in as return values (to hold
 * returned values) and as arguments (to hold input arguments) to methods
 * invoked using the Invoke method of the Method class.  This union consists
 * of all possible argument types.
 ************************************************************************** **/
typedef union Value
{
    bool bool_v;                               /**< boolean value */
    signed char char_v;                 /**< signed char value */
    unsigned char unsigned_char_v;             /**< unsigned char value */
    wchar_t wchar_v;                           /**< wchar_t value */
    signed short short_v;               /**< signed short value */
    unsigned short unsigned_short_v;           /**< unsigned short value */
    signed int int_v;                   /**< signed int value */
    unsigned int unsigned_int_v;               /**< unsigned int value */
    signed long long_v;                 /**< signed long value */
    unsigned long unsigned_long_v;             /**< unsigned long value */
    signed long long long_long_v;       /**< signed long long value */
    unsigned long long unsigned_long_long_v;   /**< unsigned long long value */
    float float_v;                             /**< float value */
    double double_v;                           /**< double value */
    long double long_double_v;                 /**< long double value */
    void *pointer_v;                           /**< pointer value */
}
Value;


/** **************************************************************************
 * An Xrtti::Context is either a namespace or a structure (class, struct,
 * or union), or is incomplete; it has a name and a parent context.
 ************************************************************************** **/
class Context
{
public:

    /** **********************************************************************
     * This enumerates the valid Context types
     ********************************************************************** **/
    enum Type
    {
        Type_Class          = 0,
        Type_Incomplete     = 1,
        Type_Namespace      = 2,
        Type_Struct         = 3,
        Type_Union          = 4
    };

    /** 
     * Returns the type of this context.  The most derived class of this
     * object is determined by the value returned here:
     *
     * Type_Class: most derived type is Class <br>
     * Type_Incomplete: most derived type is Incomplete <br>
     * Type_Namespace: most derived type is Namespace <br>
     * Type_Struct: most derived type is Struct <br>
     * Type_Union: most derived type is Union <br>
     *
     * @return the type of this context
     **/
    virtual Type GetType() const = 0;

    /**
     * Returns the name of this context.
     *
     * @return the name of this context.
     **/
    const char *GetName() const
    {
        return pNameM;
    }

    /**
     * Returns the full name of this context.  This includes the entire
     * namespace leading up to this context.  It is the name of the parent
     * Context prepended to the name of this Context, with a double-colon
     * separator.
     *
     * @return the full name of this context
     **/
    const char *GetFullName() const
    {
        return pFullNameM;
    }

    /**
     * This is the parent context of this context.  If this is NULL, then this
     * Context is the root Context and has no parent.
     **/
    const Context *GetParent() const
    {
        return pParentM;
    }

private:

    friend class Incomplete;
    friend class Namespace;
    friend class Structure;

    // Only the Xrtti library may create, destroy, or copy Contexts
    Context() { }
    virtual ~Context() { }
    Context(const Context &);
    Context &operator =(const Context &);

    const char *pNameM;

    const char *pFullNameM;
    
    Context *pParentM;
};


/** **************************************************************************
 * Incomplete represents an incomplete class, struct, or union.  An incomplete
 * type has no information other than its name, full name, and context - all
 * provided by Context.  Incomplete class, struct, and unions are those
 * referenced by name in the C++ code but never given a definition.  Nothing
 * can be discerned about such a type except for the most basic information.
 *************************************************************************** **/
class Incomplete : public Context
{
public:

    /** 
     * Returns the type of this Context, which is Type_Incomplete.
     * 
     * @return the type of this Context, which is Type_Incomplete.
     **/
    virtual Type GetType() const
    {
        return Type_Incomplete;
    }

private:

    // Only the Xrtti library may create, destroy, or copy Incompletes
    Incomplete() { }
    virtual ~Incomplete() { }
    Incomplete(const Incomplete &);
    Incomplete &operator =(const Incomplete &);
};


/** **************************************************************************
 * Namespace describes a C++ namespace.  A Namespace has no members beyond
 * those of a Context.
 ************************************************************************** **/
class Namespace : public Context
{
public:

    /** 
     * Returns the type of this Context, which is Type_Namespace.
     * 
     * @return the type of this Context, which is Type_Namespace.
     **/
    virtual Type GetType() const
    {
        return Type_Namespace;
    }

private:

    // Only the Xrtti library may create, destroy, or copy Namespaces
    Namespace() { }
    ~Namespace() { }
    Namespace(const Namespace &);
    Namespace &operator =(const Namespace &);
};


/** **************************************************************************
 * Structure is the base class for Class, Struct, and Union.  It defines the
 * members that are shared by these three context types.
 ************************************************************************** **/
class Structure : public Context
{
public:

    /**
     * Returns the access type of the Struct within the context in which it is
     * defined.
     *
     * @return the access type of the Struct within the context in which it is
     *         defined.
     **/
    AccessType GetAccessType() const
    {
        return accessTypeM;
    }

    /**
     * If the -r option was specified to the xrttigen program, then C++ RTTI
     * type_info objects will be associated with each Structure.  This method
     * will then return the type_info object associated with the C++
     * class/struct/union described by this Structure.
     *
     * If -r was not used with xrttigen, then C++ RTTI is not supported, and
     * this method will return NULL.
     *
     * @return the C++ RTTI type_info object associated with the C++ class/
     *         struct/union described by this Structure, or NULL if RTTI
     *         options were not enabled for xrttigen when it produced the
     *         Generated object resulting in this Structure.
     **/
    const std::type_info *GetTypeInfo() const
    {
        return pTypeInfoM;
    }

    /**
     * This is the number of base structures of this Structure.  Base
     * structures are, in essential, base classes, although they are not
     * necessarily classes; they may be structs or unions as well.  The
     * Base class identifies both the base Structure, and the offset from
     * this structure to the base class.
     *
     * @return the number of base structures of this Structure
     **/
    U32 GetBaseCount() const
    {
        return baseCountM;
    }

    /**
     * Returns a base structure of this Structure.
     * 
     * @param index is the number of the base Structure to return
     * @return a base structure of this Structure.
     **/
    const Base &GetBase(U32 index) const;

    /**
     * Returns the number of Fields in this Context.
     *
     * @return the number of Fields in this Context.
     **/
    U32 GetFieldCount() const
    {
        return fieldCountM;
    }

    /**
     * Returns a Field of this Context.  These are in the same order as in
     * defined in the C++ class described by this Context.  The number of
     * Fields in this Context is given by the return value of GetFieldCount().
     *
     * @param index is the number of the Field to return
     * @return a Field of this Context.
     **/
    const Field &GetField(U32 index) const;

    /**
     * Returns true if the Structure is anonymous.  Anonymous structures have
     * no type name and thus cannot be instantiated.  Therefore, the
     * Constructors will not be invokeable; and the Create and CreateArray
     * methods will return NULL.
     * 
     * @return true if the Structure is anonymous; false if not.
     **/
    bool IsAnonymous() const
    {
        return isAnonymousM;
    }

    /**
     * This is the number of constructors of this Structure.
     *
     * @return the number of constructors of this Structure
     **/
    U32 GetConstructorCount() const
    {
        return constructorCountM;
    }

    /**
     * Returns a constructor of this Structure.
     *
     * @param index is the number of the Constructor to return
     * @return a constructor of this Structure.
     **/
    const Constructor &GetConstructor(U32 index) const;

    /**
     * Returns true if this Structure has a destructor, false if not.
     *
     * @return true if this Structure has a destructor, false if not.
     **/
    bool HasDestructor() const
    {
        return pDestructorM ? true : false;
    }

    /**
     * Returns true if this Structure has a destructor.
     *
     * @return true if this Structure has a destructor.
     **/
    const Destructor &GetDestructor() const;

    /**
     * Returns a newly-constructed object of this Structure type; the
     * default constructor is used.  The returned object must be deleted
     * using the Delete() method.  If the Structure is abstract or anonymous,
     * returns NULL.
     *
     * @return a newly-constructed object of this Structure type; the
     *         default constructor is used.
     **/
    void *Create() const
    {
        if (pCreateMethodM) {
            return (*pCreateMethodM)();
        }
        else {
            return 0;
        }
    }

    /**
     * Returns a newly-constructed array of objects of this Structure type;
     * the default array constructor is used.  The returned array must be
     * deleted using the DeleteArray() method.  If the Structure is abstract,
     * or anonymous, returns NULL.
     *
     * @param count is the number of elements in the array to be created
     * @return a newly-created array of objects of this Structure type; the
     *         default constructor is used.
     **/
    void *CreateArray(U32 count) const
    {
        if (pCreateArrayMethodM) {
            return (*pCreateArrayMethodM)(count);
        }
        else {
            return 0;
        }
    }

    /**
     * Deletes an object created by the Create() method.
     *
     * @param pInstance is the object to be deleted.
     **/
    void Delete(void *pInstance) const
    {
        if (pDeleteMethodM) {
            (*pDeleteMethodM)(pInstance);
        }
    }

    /**
     * Deletes an array of objects created by the CreateArray() method.
     *
     * @param pInstanceArray is the array of objects to be deleted.
     **/
    void DeleteArray(void **pInstanceArray) const
    {
        if (pDeleteArrayMethodM) {
            (*pDeleteArrayMethodM)(pInstanceArray);
        }
    }

private:

    friend class Struct;
    friend class Union;
    
    // Only the Xrtti library may create, destroy, or copy Structures
    Structure() { }
    ~Structure() { }
    Structure(const Structure &);
    Structure &operator =(const Structure &);

    AccessType accessTypeM;

    const std::type_info *pTypeInfoM;

    U32 baseCountM;

    Base *pBasesM;

    U32 fieldCountM;
    
    class Field *pFieldsM;

    bool isAnonymousM;

    U32 constructorCountM;

    Constructor *pConstructorsM;

    Destructor *pDestructorM;
    
    void *(*pCreateMethodM)();

    void *(*pCreateArrayMethodM)(U32);

    void (*pDeleteMethodM)(void *);

    void (*pDeleteArrayMethodM)(void *);
};


/** **************************************************************************
 * Base describes a super class (or super struct or super union) of a
 * class, struct, or union.  It gives both the Structure that is the base
 * class, and also the offset from the sub class to the super class; this
 * is the value which, when added to a pointer to a sub class, yields a
 * pointer to the base class.
 ************************************************************************** **/
class Base
{
public:

    /**
     * Returns the structure which is the base class/struct/union.
     *
     * @return the structure which is the base class/struct/union.
     **/
    const Structure &GetStructure() const
    {
        return *pStructureM;
    }
    
    /**
     * Returns the offset from an instance of the subclass, to the instance
     * of the superclass.
     *
     * class Foo { };
     * class Bar : public Foo { };
     *
     * Foo *pF;
     * Bar *pB = (Bar *) ((void *) pF + offset);
     *
     * NOTE that this value will be 0 when the Base is the base of an
     * anonymous Structure.
     *
     * @return the offset from an instance of the subclass, to the instance
     *         of the superclass
     **/
    U32 GetOffset() const
    {
        return offsetM;
    }

private:

    // Only the Xrtti library may create, destroy, or copy Bases
    Base() { }
    ~Base() { }
    Base(const Base &);
    Base &operator =(const Base &);

    Structure *pStructureM;

    U32 offsetM;
};


/** **************************************************************************
 * A Union describes a C++ union.  It has no additional fields beyond that of
 * a Structure, but each of its Field's space is shared with all other Fields;
 * thus, only one Field of the union may be used without interfering with the
 * others.
 ************************************************************************** **/
class Union : public Structure
{
public:

    /** 
     * Returns the type of this Context, which is Type_Union.
     * 
     * @return the type of this Context, which is Type_Union.
     **/
    virtual Type GetType() const
    {
        return Type_Union;
    }

private:

    // Only the Xrtti library may create, destroy, or copy Unions
    Union() { }
    ~Union() { }
    Union(const Union &);
    Union &operator =(const Union &);
};


/** **************************************************************************
 * A Struct represents a C++ struct.  It is a Structure with Methods.
 ************************************************************************** **/
class Struct : public Structure
{
public:

    /** 
     * Returns the type of this Context, which is Type_Struct.
     * 
     * @return the type of this Context, which is Type_Struct.
     **/
    virtual Type GetType() const
    {
        return Type_Struct;
    }

    /**
     * Returns true if this Structure is abstract (defines at least one
     * pure virtual method), false if not.
     *
     * @return true if this Structure is abstract (defines at least one
     *         pure virtual method), false if not.
     **/
    bool IsAbstract() const
    {
        return isAbstractM;
    }

    /**
     * Returns the number of Methods of this Struct.
     *
     * @return the number of Methods of this Struct.
     **/
    U32 GetMethodCount() const
    {
        return methodCountM;
    }

    /**
     * Returns a Method of this Struct.
     *
     * @param index is the number of the Method to return
     * @return a Method of this Struct.
     **/
    const Method &GetMethod(U32 index) const;
    
private:

    friend class Class;

    // Only the Xrtti library may create, destroy, or copy Structs
    Struct() { }
    ~Struct() { }
    Struct(const Struct &);
    Struct &operator =(const Struct &);

    bool isAbstractM;

    U32 methodCountM;

    Method *pMethodsM;
};


/** **************************************************************************
 * A Class represents a C++ class.  It is no different from a Struct.
 ************************************************************************** **/
class Class : public Struct
{
public:

    /** 
     * Returns the type of this Context, which is Type_Class.
     * 
     * @return the type of this Context, which is Type_Class.
     **/
    virtual Type GetType() const
    {
        return Type_Class;
    }

private:
    
    // Only the Xrtti library may create, destroy, or copy Classes
    Class() { }
    ~Class() { }
    Class(const Class &);
    Class &operator =(const Class &);
};


/** **************************************************************************
 * Member is the base class of all members of a Structure, including
 * Constructor, Method, and Field.
 ************************************************************************** **/
class Member
{
public:

    /**
     * Returns the access type of the member.
     *
     * @return the access type of the member.
     **/
    AccessType GetAccessType() const
    {
        return accessTypeM;
    }

    /**
     * Returns the Context in which the Member is declared; it is the Context
     * which "owns" the Member and in which the Member lives.
     *
     * @return the Context in which the Member is declared
     **/
    const Context &GetContext() const
    {
        return *pContextM;
    }

    /**
     * Returns the name of this member.
     *
     * @return the name of this member.
     **/
    const char *GetName() const
    {
        return pNameM;
    }

    /**
     * Returns true if the Member is static in scope, false if it is not.
     *
     * @return true if the Member is static in scope, false if it is not.
     **/
    bool IsStatic() const
    {
        return isStaticM;
    }

private:

    friend class Constructor;
    friend class Destructor;
    friend class Field;
    friend class Method;

    // Only the Xrtti library may create, destroy, or copy Members
    Member() { }
    ~Member() { }
    Member(const Member &);
    Member &operator =(const Member &);

    AccessType accessTypeM;

    Context *pContextM;

    const char *pNameM;

    bool isStaticM;
};


/** **************************************************************************
 * Field represents a C++ field (i.e. member variable).
 ************************************************************************** **/
class Field : public Member
{
public:

    /**
     * Returns the type of the Field.
     *
     * @return the type of the Field.
     **/
    const Type &GetType() const
    {
        return *pTypeM;
    }

    /**
     * Returns the offset in bytes from the beginning of the containing
     * Structure to this Field.  This field's value is undefined for all
     * static Fields, and for all bitfield Fields.  NOTE that this method
     * returns 0 for Fields of anonymous Structures.
     *
     * @return the offset in bytes from the beginning of the containing
     *         Structure to this Field, or 0 if the Field is a field of an
     *         anonymous Structure
     **/
    U32 GetOffset() const
    {
        return offsetM;
    }

    /**
     * Returns the current value of this Field within the given instance,
     * which is an object of the type of the context containing this field.
     * Returns an undefined value if the containing class is anonymous or the
     * field is an anonymous structure.
     *
     * @param pInstance is an object of the type of the context containing
     *        this field.  NULL may be passed in if the field is static.
     * @param returnValue will return the value of this Field within the given
     *         instance.  See the documentation for the Value union for an
     *         explanation of which union member will be set.
     **/
    void Get(void *pInstance, Value &returnValue) const
    {
        (*pGetMethodM)(pInstance, returnValue);
    }

    /**
     * Sets the current value of this Field within the given instance, which
     * is an object of the type of the context containing this field.  Has no
     * effect if the containing class is anonymous.  This method must not be
     * called for static Fields.
     *
     * @param pInstance is an object of the type of the context containing
     *        this field.  NULL may be passed in if the field is static.
     * @param value holds the value of this Field to set within the given
     *        instance.  See the documentation for the Value union for an
     *        explanation of which union member to set.
     **/
    void Set(void *pInstance, const Value &value) const
    {
        (*pSetMethodM)(pInstance, value);
    }

private:

    // Only the Xrtti library may create, destroy, or copy Fields
    Field() { }
    ~Field() { }
    Field(const Field &);
    Field &operator =(const Field &);

    Type *pTypeM;

    U32 offsetM;

    void (*pGetMethodM)(void *, Value &);

    void (*pSetMethodM)(void *, const Value &);
};


/** **************************************************************************
 * Argument represents a single argument in a constructor or function
 * signature.
 ************************************************************************** **/
class Argument
{
public:

    /**
     * Returns the type of the argument.
     *
     * @return the type of the argument.
     **/
    const Type &GetType() const
    {
        return *pTypeM;
    }

    /**
     * Returns true if the Argument has a default value, false if not.
     *
     * @return true if the Argument has a default value, false if not.
     **/
    bool HasDefault() const
    {
        return hasDefaultM;
    }

    /**
     * If HasDefault() returns true, then this returns the default value for
     * this argument.  If HasDefault() returns false, then this returns an
     * undefined value.
     * 
     * @return the default value for this argument.
     **/
    const Value &GetDefault() const
    {
        return defaultM;
    }
     
private:

    // Only the Xrtti library may create, destroy, or copy Arguments
    Argument() { }
    ~Argument() { }
    Argument(const Argument &);
    Argument &operator =(const Argument &);

    Type *pTypeM;

    bool hasDefaultM;

    Value defaultM;
};


/** **************************************************************************
 * DestructorSignature describes the signature of a destructor
 ************************************************************************** **/
class DestructorSignature
{
public:

    /**
     * Returns the number of types which this destructor declares that it
     * throws as exceptions.
     *
     * @return the number of types which this destructor declares that it
     *         throws as exceptions.
     **/
    U32 GetThrowCount() const
    {
        return throwCountM;
    }

    /**
     * Returns a type which this destructor declares that it throws as an
     * exception.
     *
     * @param index is the index of the throw type to return
     * @return a type which this destructor declares that it throws as an
     *         exception.
     **/
    const Type &GetThrow(U32 index) const
    {
        return *(pThrowsM[index]);
    }

private:

    friend class Destructor;
    friend class ConstructorSignature;

    // Only the Xrtti library may create, destroy, or copy DestructorSignature
    // objects
    DestructorSignature() { }
    ~DestructorSignature() { }
    DestructorSignature(const DestructorSignature &);
    DestructorSignature &operator =(const DestructorSignature &);

    U32 throwCountM;
    Type **pThrowsM;
};


/** **************************************************************************
 * ConstructorSignature describes the signature of a constructor
 ************************************************************************** **/
class ConstructorSignature : public DestructorSignature
{
public:

    /**
     * Returns the number of arguments of the constructor.
     *
     * @return the number of arguments of the constructor.
     **/
    U32 GetArgumentCount() const
    {
        return argumentCountM;
    }

    /**
     * Returns the argument at index [index].
     *
     * @return the argument at index [index].
     **/
    const Argument &GetArgument(U32 index) const
    {
        return pArgumentsM[index];
    }

    /**
     * Returns true if the constructor has ellipsis - an undefined number of
     * arguments following the final declared argument, or false if it
     * does not.
     *
     * @return true if the constructor has ellipsis
     **/
    bool HasEllipsis() const
    {
        return hasEllipsisM;
    }

private:

    friend class Constructor;
    friend class MethodSignature;

    // Only the Xrtti library may create, destroy, or copy ConstructorSignature
    // objects
    ConstructorSignature() { }
    ~ConstructorSignature() { }
    ConstructorSignature(const ConstructorSignature &);
    ConstructorSignature &operator =(const ConstructorSignature &);

    U32 argumentCountM;

    Argument *pArgumentsM;

    bool hasEllipsisM;
};


/** **************************************************************************
 * MethodSignature describes the signature of a function
 ************************************************************************** **/
class MethodSignature : public ConstructorSignature
{
public:

    /**
     * Returns the return type of the function.
     *
     * @return the return type of the function.
     **/
    const Type &GetReturnType() const
    {
        return *pReturnTypeM;
    }

private:

    friend class Method;
    friend class TypeFunction;

    // Only the Xrtti library may create, destroy, or copy MethodSignature
    // objects
    MethodSignature() { }
    ~MethodSignature() { }
    MethodSignature(const MethodSignature &);
    MethodSignature &operator =(const MethodSignature &);

    Type *pReturnTypeM;
};


/** **************************************************************************
 * Destructor represents a C++ destructor
 ************************************************************************** **/
class Destructor : public Member
{
public:

    /**
     * Returns true if the destructor is virtual, false if it is concrete.
     *
     * @return true if the destructor is virtual, false if it is concrete.
     **/
    bool IsVirtual() const
    {
        return isVirtualM;
    }

    /**
     * Returns true if the destructor is pure virtual, false if it is not.
     *
     * @return true if the destructor is pure virtual, false if it is not.
     **/
    bool IsPureVirtual() const
    {
        return isPureVirtualM;
    }

    /**
     * Returns an object describing the signature of this destructor.
     *
     * @return an object describing the signature of this destructor.
     **/
    const DestructorSignature &GetSignature() const
    {
        return signatureM;
    }

    /**
     * Invokes this destructor on a constructed object.
     * constructed instance must be deleted by the caller.  If the containing
     * Structure is anonymous, this method will return NULL, since anonymous
     * Structures cannot be instantiated.
     *
     * @param pInstance is the instance of the object to destroy
     **/
    void Invoke(void *pInstance) const
    {
        (*pInvokeDestructorM)(pInstance);
    }

private:

    // Only the Xrtti library may create, destroy, or copy Destructors
    Destructor() { }
    ~Destructor() { }
    Destructor(const Destructor &);
    Destructor &operator =(const Destructor &);

    bool isVirtualM, isPureVirtualM;

    DestructorSignature signatureM;

    void (*pInvokeDestructorM)(void *);
};


/** **************************************************************************
 * Constructor represents a C++ constructor.
 ************************************************************************** **/
class Constructor : public Member
{
public:

    /**
     * Returns an object describing the signature of this constructor.
     *
     * @return an object describing the signature of this constructor.
     **/
    const ConstructorSignature &GetSignature() const
    {
        return signatureM;
    }

    /**
     * Additionally, a Constructor gives names to its arguments.  This returns
     * the name of argument at index [index].
     *
     * @return the name of argument at index [index]
     **/
    const char *GetArgumentName(U32 index) const
    {
        return pArgumentNamesM[index];
    }

    /**
     * Invokes this constructor and returns the constructed instance.  The
     * constructed instance must be deleted by the caller.  If the containing
     * Structure is anonymous, this method will return NULL, since anonymous
     * Structures cannot be instantiated.
     *
     * @param pArgumentValues is an array of Values holding the arguments to
     *        this constructor.  The specific union member which will be used
     *        is the one whose value is the same type as the argument value
     *        of the constructor.  The number of elements in pArgumentValues
     *        must match the number of arguments to this constructor.
     * @return a newly-constructed instance of the containing Structure; or
     *         NULL if the containing Structure is anonymous
     **/
    void *Invoke(Value *pArgumentValues) const
    {
        Value ret;
        (*pInvokeConstructorM)(0, /* returns */ ret, pArgumentValues);
        return ret.pointer_v;
    }

private:

    // Only the Xrtti library may create, destroy, or copy Constructors
    Constructor() { }
    ~Constructor() { }
    Constructor(const Constructor &);
    Constructor &operator =(const Constructor &);

    ConstructorSignature signatureM;

    const char **pArgumentNamesM;
    
    void (*pInvokeConstructorM)(void *, Value &, Value *);
};


/** **************************************************************************
 * Method represents a C++ method.  
 ************************************************************************** **/
class Method : public Member
{
public:

    /**
     * Returns true if the method is const, false if it is not.
     *
     * @return true if the method is const, false if it is not.
     **/
    bool IsConst() const
    {
        return isConstM;
    }

    /**
     * Returns true if the method is virtual, false if it is concrete.
     *
     * @return true if the method is virtual, false if it is concrete.
     **/
    bool IsVirtual() const
    {
        return isVirtualM;
    }

    /**
     * Returns true if the method is pure virtual, false if it is not.
     *
     * @return true if the method is pure virtual, false if it is not.
     **/
    bool IsPureVirtual() const
    {
        return isPureVirtualM;
    }

    /**
     * Returns an object describing the signature of this method.
     *
     * @return an object describing the signature of this method.
     **/
    const MethodSignature &GetSignature() const
    {
        return signatureM;
    }

    /**
     * Additionally, a Method gives names to its arguments.  This returns
     * the name of argument at index [index].
     *
     * @return the name of argument at index [index]
     **/
    const char *GetArgumentName(U32 index) const
    {
        return pArgumentNamesM[index];
    }

    /**
     * Invokes this method on an instance of the class represented by the
     * Class containing this Method.  It is an error to attempt to invoke
     * a pure virtual method.  This method has no effect if the containing
     * Structure is anonymous, since methods cannot be called on anonymous
     * structures.
     *
     * @param pInstance is the instance of the class to invoke this method on
     * @param returnValue will return the return value resulting from the
     *        invocation of the method.  If the method has no return value,
     *        then this argument is ignored.  The specific union member which
     *        will be set is the one whose value is the same type as the
     *        return value of the method.  If the return value is a pointer,
     *        then it is pointer_v which is set.
     * @param pArgumentValues is an array of Values holding the arguments to
     *        this method.  The specific union member which will be used is
     *        the one whose value is the same type as the argument value of
     *        the method.  The number of elements in pArgumentValues must
     *        match the number of arguments to this method.
     **/
    void Invoke(void *pInstance, Value &returnValue, 
                Value *pArgumentValues) const
    {
        (*pInvokeMethodM)(pInstance, returnValue, pArgumentValues);
    }

private:

    // Only the Xrtti library may create, destroy, or copy Methods
    Method() { }
    ~Method() { }
    Method(const Method &);
    Method &operator =(const Method &);

    bool isConstM, isVirtualM, isPureVirtualM;

    MethodSignature signatureM;

    const char **pArgumentNamesM;

    void (*pInvokeMethodM)(void *, Value &, Value *);
};


/** **************************************************************************
 * Type represents a C++ type.  Each Type may have a number of Pointers,
 * each of which describes the properties of one level of pointer indirection
 * of the Type, and a number of Arrays, each of which describes the properties
 * of one array dimension.  For example, the C++ type:
 *
 * const * volatile * int[5][];
 *
 * ... would have two Pointers (the first one volatile, the second const), and
 * two Arrays (the first one with element count 5, the second with element
 * count unbounded).
 ************************************************************************** **/
class Type
{
public:

    /** **********************************************************************
     * These are the C++ types which may be represented by a Type object.
     *
     * If the Type of this Type is BaseType_Enumeration, then this Type is
     * actually a TypeEnumeration object.
     *
     * If the Type of this Type is BaseType_Function, then this Type is
     * actually a TypeFunction object.
     *
     * If the Type of this Type is BaseType_Structure, then this Type is
     * actually a TypeStructure object.
     *
     * Otherwise, the Type object is not subclassed.
     ********************************************************************** **/
    enum BaseType
    {
        BaseType_Void                    = 0,
        BaseType_Bool                    = 1,
        BaseType_Char                    = 2,
        BaseType_Unsigned_Char           = 3,
        BaseType_WChar                   = 4,
        BaseType_Short                   = 5,
        BaseType_Unsigned_Short          = 6,
        BaseType_Int                     = 7,
        BaseType_Unsigned_Int            = 8,
        BaseType_Long                    = 9,
        BaseType_Unsigned_Long           = 10,
        BaseType_Long_Long               = 11,
        BaseType_Unsigned_Long_Long      = 12,
        BaseType_Float                   = 13,
        BaseType_Double                  = 14,
        BaseType_Long_Double             = 15,
        BaseType_Bitfield                = 16,
        BaseType_Enumeration             = 17,
        BaseType_Function                = 18,
        BaseType_Structure               = 19
    };

    /**
     * Returns the C++ type of this Type.
     *
     * @return the C++ type of this Type.
     **/
    BaseType GetBaseType() const
    {
        return baseTypeM;
    }

    /**
     * If this returns true, then the C++ type represented by this Type
     * is const.  If this returns false, then it is not.
     *
     * @return true if C++ type represented by this Type is const
     **/
    bool IsConst() const
    {
        return isConstM;
    }

    /**
     * If this returns true, then the C++ type represented by this Type
     * is volatile.  If this returns false, then it is not.
     *
     * @return true if C++ type represented by this Type is volatile
     **/
    bool IsVolatile() const
    {
        return isVolatileM;
    }

    /**
     * If this returns true, then the C++ type represented by this Type
     * is a C++ reference.  If this returns false, then it is not.
     *
     * @return true if C++ type represented by this Type is a C++ reference
     **/
    bool IsReference() const
    {
        return isReferenceM;
    }

    /**
     * If the type is Type_Bitfield, then this will return the number of bits
     * in the bitfield.
     *
     * @return the number of bits in the bitfield if the type is Type_Bitfield
     **/
    U32 GetBitfieldBitCount() const
    {
        return bitCountM;
    }

    /**
     * Returns the pointer "depth" of this Type, which is the number of
     * times that this type must be dereferenced to get to a value of the
     * basic C++ type of this Type.
     *
     * @return the pointer "depth" of this Type
     **/
    U32 GetPointerCount() const
    {
        return pointerCountM;
    }

    /**
     * Returns a description of the dereference characteristics of the
     * given pointer depth for this Type.
     *
     * @param depth is the pointer depth to get the dereference
     *              characteristics of
     * @return a description of the dereference characteristics of the
     *         given pointer depth for this Type.
     **/
    const Pointer &GetPointer(U32 depth) const;

    /**
     * Returns the number of array dimensions of this Type.
     *
     * @return the number of array dimensions of this Type.
     **/
    U32 GetArrayCount() const
    {
        return arrayCountM;
    }

    /**
     * Returns a description of the number of elements of the given array
     * dimension of this Type.
     *
     * @param dimension is the array dimension of this Type to get the number
     *        of elements of
     * @return the number of elements of the given array dimension of this
     *        Type
     **/
    const Array &GetArray(U32 dimension) const;

private:

    friend class TypeEnumeration;
    friend class TypeFunction;
    friend class TypeStructure;

    // Only the Xrtti library may create, destroy, or copy Types
    Type() { }
    ~Type() { }
    Type(const Type &);
    Type &operator =(const Type &);

    BaseType baseTypeM;

    bool isConstM;

    bool isVolatileM;

    bool isReferenceM;

    U32 bitCountM;

    U32 pointerCountM;

    Pointer *pPointersM;

    U32 arrayCountM;

    Array *pArraysM;
};


/** **************************************************************************
 * Pointer describes the dereference characteristics of a single pointer depth
 *  of a Type.
 ************************************************************************** **/
class Pointer
{
public:

    /**
     * If this returns true, then the dereferenced value is const.  If this
     * returns false, it is not.
     *
     * @return true if the dereferenced value is const
     **/
    bool IsConst() const
    {
        return isConstM;
    }

    /**
     * If this returns true then the dereferenced value is volatile.  If this
     * returns false, it is not.
     *
     * @return true if the dereferenced value is volatile
     **/
    bool IsVolatile() const
    {
        return isVolatileM;
    }

private:

    // Only the Xrtti library may create, destroy, or copy Pointers
    Pointer() { }
    ~Pointer() { }
    Pointer(const Pointer &);
    Pointer &operator =(const Pointer &);

    bool isConstM;

    bool isVolatileM;
};


/** **************************************************************************
 * Array describes the element count of a single array dimension of a Type.
 ************************************************************************** **/
class Array
{
public:

    /**
     * If this returns true, the array dimension is of unbounded size; it may
     * have any element count.  If this returns false, the element count is
     * fixed and specified, and can be obtained by calling GetElementCount().
     *
     * @return true if the array dimension is of unbounded size
     **/
    bool IsUnbounded() const
    {
        return isUnboundedM;
    }

    /**
     * If the array dimension is not unbounded, then this specifies the
     * element count of the array dimension.  The index of the last element of
     * the of the array is at this value minus one.
     *
     * @return the element count of the array dimension
     **/
    U32 GetElementCount() const 
    {
        return elementCountM;
    }

private:

    // Only the Xrtti library may create, destroy, or copy Arrays
    Array() { }
    ~Array() { }
    Array(const Array &);
    Array &operator =(const Array &);

    bool isUnboundedM;

    U32 elementCountM;
};


/** **************************************************************************
 * Enumeration describes a C++ enum.
 ************************************************************************** **/
class Enumeration
{
public:

    /**
     * Returns the access type of the Enuimeration within the context in which
     * it is defined.
     *
     * @return the access type of the Enumeration within the context in which
     *         it is defined.
     **/
    AccessType GetAccessType() const
    {
        return accessTypeM;
    }

    /**
     * Returns the Context in which this Enumeration is defined.
     *
     * @return the Context in which this Enumeration is defined
     **/
    const Context &GetContext() const
    {
        return *pContextM;
    };

    /**
     * Returns the name of this Enumeration.
     *
     * @return the name of this Enumeration
     **/
    const char *GetName() const
    {
        return pNameM;
    }
    
    /**
     * Returns the number of EnumerationValues of this Enumeration.
     *
     * @return the number of EnumerationValues of this Enumeration.
     **/
    U32 GetValueCount() const
    {
        return valueCountM;
    }

    /**
     * Returns an EnumerationValue of this Enumeration.
     *
     * @param index is the number of the EnumerationValue to return
     * @return an EnumerationValue of this Enumeration
     **/
    const EnumerationValue &GetValue(U32 index) const;

private:

    // Only the Xrtti library may create, destroy, or copy Enumerations
    Enumeration() { }
    ~Enumeration() { }
    Enumeration(const Enumeration &);
    Enumeration &operator =(const Enumeration &);

    AccessType accessTypeM;

    Context *pContextM;

    const char *pNameM;

    U32 valueCountM;

    EnumerationValue *pValuesM;
};


/** **************************************************************************
 * EnumerationValue represents a single value of a C++ enum.
 ************************************************************************** **/
class EnumerationValue
{
public:

    /**
     * Returns the name of the value.
     *
     * @return the name of the value.
     **/
    const char *GetName() const
    {
        return pNameM;
    }

    /**
     * Returns the numerical const value of the value.
     *
     * @return the numerical const value of the value
     **/
    S32 GetValue() const
    {
        return valueM;
    }

private:

    // Only the Xrtti library may create, destroy, or copy EnumerationValues
    EnumerationValue() { }
    ~EnumerationValue() { }
    EnumerationValue(const EnumerationValue &);
    EnumerationValue &operator =(const EnumerationValue &);

    const char *pNameM;

    S32 valueM;
};


/** **************************************************************************
 * TypeEnumeration represents a C++ type that is an enum.
 ************************************************************************** **/
class TypeEnumeration : public Type
{
public:

    /**
     * This returns the Enumeration of the TypeEnumeration, which describes
     * the set of allowed enumerated values that values of this type may
     * have.
     *
     * @return the set of allowed enumerated values for this type
     **/
    const Enumeration &GetEnumeration() const
    {
        return *pEnumerationM;
    }

private:

    // Only the Xrtti library may create, destroy, or copy TypeEnumerations
    TypeEnumeration() { }
    ~TypeEnumeration() { }
    TypeEnumeration(const TypeEnumeration &);
    TypeEnumeration &operator =(const TypeEnumeration &);

    Enumeration *pEnumerationM;
};


/** **************************************************************************
 * TypeEnumeration represents a C++ type that is a function.
 ************************************************************************** **/
class TypeFunction : public Type
{
public:

    /**
     * Returns an object describing the signature of the function type.
     *
     * @return an object describing the signature of the function type.
     **/
    const MethodSignature &GetSignature() const
    {
        return signatureM;
    }

private:

    // Only the Xrtti library may create, destroy, or copy TypeFunctions
    TypeFunction() { }
    ~TypeFunction() { }
    TypeFunction(const TypeFunction &);
    TypeFunction &operator =(const TypeFunction &);
    
    MethodSignature signatureM;
};


/** **************************************************************************
 * TypeStructure represents a C++ type which is a structure (class, struct,
 * or union).
 ************************************************************************** **/
class TypeStructure : public Type
{
public:

    /**
     * Returns the Structure defining the type.
     *
     * @return the Structure defining the type.
     **/
    const Structure &GetStructure() const
    {
        return *pStructureM;
    }

private:

    // Only the Xrtti library may create, destroy, or copy TypeFunctions
    TypeStructure() { }
    ~TypeStructure() { }
    TypeStructure(const TypeStructure &);
    TypeStructure &operator =(const TypeStructure &);

    Structure *pStructureM;
};


/** **************************************************************************
 * Xrtti Methods
 ************************************************************************** **/

/**
 * Registers a set of objects generated by the xrttigen program.  This may be
 * called multiple times with the output of multiple runs of xrttigen to build
 * up a complete set of registered Xrtti objects.
 *
 * Running xrttigen against your program's header files, and then registering
 * the resulting Generated object with this method, is the means by which the
 * xrtti system gets information about your classes.  For each class
 * registered in this way, you can find the class in the list of Contexts
 * using GetContext(), or look it up by full name, using LookupContext().
 *
 * @param pGenerated is the set of objects generated by the xrttigen program.
 **/
void Register(const Generated *pGenerated);

/**
 * Returns the number of registered Contexts.
 * 
 * @return the number of registered Contexts.
 **/
U32 GetContextCount();

/**
 * Returns a Context.  The total number of available contexts is available via
 * a call to GetContextCount().
 *
 * @param index is the index of the Context to get; this value must be less
 *        than the number of Contexts returned by GetContextCount().
 * @return the index'th Context. 
 **/
const Context *GetContext(U32 index);

/**
 * Looks up a Context by full name, and returns it.  Returns NULL if there
 * is no such registered Context.
 *
 * @param pFullName is the full name (i.e. foo::bar::baz for C++, or just
 *        the struct/union name for C) of the Context to look up and return
 * @return the Context with the given full name, or NULL if there is no
 *         such Context
 **/
const Context *LookupContext(const char *pFullName);

/**
 * Looks up a Structure by C++ RTTI type_info, and returns it.  This only
 * works if the Structure was generated by xrttigen with the -r option
 * enabled.  Returns NULL if there is no such registered Structure.
 *
 * @param typeinfo is the C++ RTTI type_info to fetch the corresponding
 *        Structure for.  Only works if xrttigen had the -r option specified.
 * @return the Structure with the given type_info, or NULL if there is no
 *         such Structure
 **/
const Structure *LookupStructure(const std::type_info &typeinfo);

/**
 * This may be called to shut down the Xrtti system.  It will destroy all
 * Xrtti objects created during the registration process.  After this function
 * is called, no more Xrtti structures will be present and the Xrtti system
 * should not be used.  The purpose of this function is primarily to assist in
 * debugging, where the termination of the program with the Xrtti objects
 * still allocated may look like a memory leak.  Calling this function is not
 * at all necessary - the objects don't need to be destroyed before the
 * program exits.
 **/
void ShutDown();

}; // namespace Xrtti

#endif // XRTTI_H
