/*****************************************************************************\
 *                                                                           *
 * XrttiPrivate.h                                                            *
 *                                                                           *
 * ------------------------------------------------------------------------- *
 * Copyright (C) 2007 Bryan Ischo <bryan@ischo.com>                          *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License Version 2 as published  *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to:                                *
 * The Free Software Foundation, Inc.                                        *
 * 51 Franklin Street, Fifth Floor                                           *
 * Boston, MA 02110-1301, USA.                                               *
 * ------------------------------------------------------------------------- *
 *                                                                           *
 * These are private data structure definitions shared between               *
 * xrttigen-generated code and the xrtti library.                            *
 *                                                                           *
 * THIS HEADER FILE IS NOT MEANT TO BE INSPECTED OR USED BY ANY END-USER OF  *
 * THE XRTTI LIBRARY.  IGNORE THE CONTENTS OF THIS FILE.                     *
 *                                                                           *
\*****************************************************************************/

#ifndef XRTTI_PRIVATE_H
#define XRTTI_PRIVATE_H

/**
 * These "cheats" enables the xrttigen-generated code to include offsetofs
 * for private and protected members and to call private and protected methods.
 **/
#define private public
#define protected public
#if 1 // Need to implement a compiler flag that turns this on and off because
      // some standard headers make this better and some make this worse
#define class struct
#endif

#include <Xrtti/Xrtti.h>

/**
 * A known address guaranteed to be in the program's memory space
 **/
#define XrttiPrivate_KnownAddress ((char *) ::Xrtti::Register)

/**
 * Really want to use real offsetof() as defined in stddef.h, but it doesn't
 * seem to work?  Why do C++ compiler writers and standard definers do such
 * stupid things???  Like just assume that nobody would ever want to use
 * offsetof for classes?
 **/
#define XrttiPrivate_offsetof(type, member) \
    (((char *) &((type *) XrttiPrivate_KnownAddress)->member) - \
     XrttiPrivate_KnownAddress)

#define XrttiPrivate_classoffsetof(subtype, type) \
    (((char *) (type *) (subtype *) XrttiPrivate_KnownAddress) - \
     XrttiPrivate_KnownAddress)

namespace XrttiPrivate {

#if 0 // This fixes indentation in emacs
}
#endif

typedef struct Context Context;
typedef struct Enumeration Enumeration;
typedef struct Field Field;
typedef struct Method Method;
typedef struct Type Type;

struct Context
{
    /* anonymous, namespace, incomplete, class, struct, union */
    ::Xrtti::Context::Type type;

    /* i.e. baz */
    const char *name;

    /* i.e. ::foo::bar::baz */
    const char *full_name;

    /* Index into the array of contexts, or 0xFFFFFFFF if no parent */
    ::Xrtti::U32 parent;

    /* Incomplete and namespace Contexts are terminated here */

    /* Access type for this context; only used for type Type_Class,
       Type_Struct, and Type_Union */
    ::Xrtti::AccessType access_type;

    /* type_info, if -r option was used in xrttigen, or NULL if not */
    const std::type_info *typeinfo;

    /* Indexes into the array of contexts */
    ::Xrtti::U32 supers_count;
    const ::Xrtti::U32 *supers;

    /* Each one is the offset of the corresponding super */
    const ::Xrtti::U32 *supers_offsets;

    /* The fields of this Context */
    ::Xrtti::U32 fields_count;
    const Field *fields;

    /* anonymous */
    bool is_anonymous;

    /* Constructors */
    ::Xrtti::U32 constructors_count;
    const Method *constructors;

    /* Destructor */
    const Method *destructor;

    /* Methods */
    ::Xrtti::U32 methods_count;
    const Method *methods;

    /* Create method */
    void *(*create_one)();

    /* Create array method */
    void *(*create_array)(::Xrtti::U32);
    
    /* Delete method */
    void (*delete_one)(void *);

    /* Delete array method */
    void (*delete_array)(void *);
};

struct Enumeration
{
    ::Xrtti::AccessType access_type;

    /* Index into the array of contexts, or 0xFFFFFFFF if no context */
    ::Xrtti::U32 context;

    const char *name;

    ::Xrtti::U32 values_count;
    const char * const *value_names;
    const ::Xrtti::S32 *value_values;
};

struct Field
{
    /* Access type for this field */
    ::Xrtti::AccessType access_type;

    const char *name;

    /* Index into the array of types */
    ::Xrtti::U32 type_index;

    bool is_static;

    ::Xrtti::U32 offset;

    /* Get method */
    void (*get)(void *, ::Xrtti::Value &);

    /* Set method */
    void (*set)(void *, const ::Xrtti::Value &);
};

struct Method
{
    /* Access type for this field */
    ::Xrtti::AccessType access_type;

    const char *name;

    /* Flags */
    bool is_static;
    bool is_const;
    bool is_virtual;
    bool is_pure_virtual;

    /* Return type index into the array of types - ignored for constructors */
    ::Xrtti::U32 return_type_index;

    /* Arguments */
    ::Xrtti::U32 arguments_count;
    const char * const *argument_names; /* ignored for function types */
    const ::Xrtti::U32 *argument_types; /* indexes into the array of types */
    void **argument_defaults;
    bool has_ellipsis;

    /* Throws */
    ::Xrtti::U32 throws_count;
    const ::Xrtti::U32 *throws; /* indexes into the array of types */

    /* Invoke method */
    void (*invoke)(void *, ::Xrtti::Value &, ::Xrtti::Value *);

};

struct Type
{
    ::Xrtti::Type::BaseType type;
    
    /* Enumeration extras */
    ::Xrtti::U32 enumeration; /* index into the array of enumerations */

    /* Function extras */
    const Method *method;

    /* Structure extras */
    ::Xrtti::U32 structure; /* index into array of contexts of a Type_Struct */

    /* Flags */
    bool is_const, is_volatile, is_reference;

    /* Bit count */
    ::Xrtti::U32 bit_count;

    /* Pointers */
    ::Xrtti::U32 pointers_count;
    const bool *pointers_const;
    const bool *pointers_volatile;

    ::Xrtti::U32 arrays_count;
    const bool *arrays_unbounded;
    const ::Xrtti::U32 *arrays_element_counts;
};

}; // namespace XrttiPrivate

namespace Xrtti {

#if 0 // This fixes indentation in emacs
}
#endif

struct Generated
{
    ::Xrtti::U32 contexts_count;
    const ::XrttiPrivate::Context * const *contexts;

    ::Xrtti::U32 enumerations_count;
    const ::XrttiPrivate::Enumeration * const *enumerations;

    ::Xrtti::U32 types_count;
    const ::XrttiPrivate::Type * const *types;
};

class AutoRegister
{
public:

    AutoRegister(const Generated &generated)
    {
        ::Xrtti::Register(&generated);
    }
};

}; // namespace Xrtti

typedef ::XrttiPrivate::Context XP_Context;
typedef ::XrttiPrivate::Enumeration XP_Enumeration;
typedef ::XrttiPrivate::Field XP_Field;
typedef ::XrttiPrivate::Method XP_Method;
typedef ::XrttiPrivate::Type XP_Type;
typedef ::Xrtti::S32 X_S32;
typedef ::Xrtti::U32 X_U32;
typedef ::Xrtti::AutoRegister X_AutoRegister;
typedef ::Xrtti::Generated X_Generated;
typedef ::Xrtti::Type X_Type;
typedef ::Xrtti::Value X_Value;

#endif /* XRTTI_PRIVATE_H */
